#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function, unicode_literals

import os
import sys
import gzip
import base64
import struct
import hashlib
import sqlite3


"""
this reproduces incorrect sqlite3 behavior under pypy3:
 * exceptions on valid input, then succeed on retry
 * truncated SELECT results

observed: PyPy 7.3.3 "Python 3.6.12" db1e853f94de42ad711bd930222bd2434e0f900d, GCC 9.2.0, gentoo
    pass: PyPy 7.3.3 "Python 2.7.18" a29ef73f9b32953753d0dd6d2a56255fa2892e24, GCC 9.2.0, gentoo
observed: 3.7.9 (00e81bd5bc52, Feb 22 2021, 02:01:13) [PyPy 7.3.4-alpha0 with MSC v.1927 64 bit (AMD64)]
observed: 3.6.12 (118564033d0a, Nov 26 2020, 02:01:17) [PyPy 7.3.4-alpha0 with MSC v.1927 32 bit]
observed: 3.6.9 (d011d093933c, May 20 2020, 01:02:03) [PyPy 7.3.2-alpha0 with MSC v.1912 32 bit]
observed: 3.6.9 (3b33cb4a5676, Apr 18 2020, 02:09:30) [PyPy 7.3.2-alpha0 with MSC v.1912 32 bit]
observed: 3.6.9 (f469952df50d, Mar 30 2020, 05:24:01) [PyPy 7.3.2-alpha0 with MSC v.1925 32 bit]
observed: 3.6.9 (2972cab4abc6, Mar 01 2020, 02:02:36) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
observed: 3.6.9 (c86c14b9e40b, Feb 05 2020, 02:02:08) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
observed: 3.6.9 (39898aeb2799, Jan 28 2020, 02:01:54) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
observed: 3.6.9 (73d20edb41e9, Jan 26 2020, 02:02:35) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
observed: 3.6.9 (73d20edb41e9, Jan 26 2020, 02:00:14) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
observed: 3.6.9 (3811ac3540f6, Jan 23 2020, 02:00:13) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
observed: 3.6.9 (55b8f216cff8, Jan 21 2020, 02:00:14) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
observed: 3.6.9 (a443ea8e15fc, Jan 17 2020, 02:00:14) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
observed: 3.6.9 (928110c124ae, Jan 14 2020, 02:00:13) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
    pass: 3.6.9 (1b19e41a3920, Jan 10 2020, 02:00:14) [PyPy 7.3.1-alpha0 with GCC 7.3.1 20180303 (Red Hat 7.3.1-5)]
    pass: 3.6.9 (8ab16affda0a, Jan 09 2020, 02:01:39) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
    pass: 3.6.9 (c092ffe0d27c, Jan 03 2020, 02:02:19) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
    pass: 3.6.9 (98943a15dbc8, Dec 26 2019, 02:02:24) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
    pass: 3.6.9 (010bdf9668e6, Dec 19 2019, 03:50:31) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]
    pass: 3.6.9 (c7b17edd132e, Dec 08 2019, 02:01:38) [PyPy 7.3.1-alpha0 with MSC v.1911 32 bit]

bisecter for msys2 or linux, takes nightly URLs:
  while true; do rm -rf pypy*; echo; read -p nightly-zip-url:; wget "$REPLY"; bin=pypy3.exe; tar -xf pypy*.bz2 && bin=bin/pypy3 || unzip pypy*.zip; result=pass; for n in {1..10}; do pypy*/$bin ../../dev/pypy3-sqlite-bug.py || { result=observed; break; }; done; printf '%8s: ' $result; pypy*/$bin -c "import sys; print(sys.version.replace('\n', ' '))"; rm -rf pypy*; done

latest version of this file:
  https://ocv.me/dev/?gist/pypy3-sqlite-bug.py
"""


FN_DB = "up2k.db"
FN_TXT = "up2k.txt.gz"


def randsrc():
    hasher = hashlib.sha1()
    buf = b"balus"
    while True:
        hasher.update(buf)
        buf = hasher.digest()
        yield buf


def make_test_data(ntotal):
    print("creating {} lines of test data in {}".format(ntotal, FN_TXT))
    rgen = randsrc()
    with gzip.open(FN_TXT, "wb") as f:
        for n in range(ntotal):
            if n % 5000 == 0:
                print(n)

            rand_ch = base64.b32encode(next(rgen))
            randtxt = rand_ch.decode("ascii")
            rd = randtxt + randtxt
            fn = randtxt[20:]
            mt, sz = struct.unpack("<LL", rand_ch[:8])

            ln = "{} {} {}/{}\n".format(mt, sz, rd, fn)
            f.write(ln.encode("utf-8"))


class Up2k(object):
    def __init__(self):
        pass

    def build_indexes(self):
        if not os.path.exists(FN_TXT):
            make_test_data(99890)

        n_added = 0
        n_total = 0
        reproduced = False
        db = self._open_db(FN_DB)
        with gzip.open(FN_TXT, "rb") as f:
            for ln in f:
                n_total += 1
                if n_total % 5000 == 0:
                    print(n_total)

                ts, sz, rp = ln.split(b" ", 2)
                ts = int(ts)
                sz = int(sz)
                rp = rp[:-1].decode("utf-8", "replace")

                rd, fn = rp.rsplit("/", 1) if "/" in rp else ["", rp]
                sql = "select * from up where rd = ? and fn = ?"
                v = (rd, fn)
                try:
                    c = db.execute(sql, v)
                except Exception as ex:
                    print("select: " + repr(ex) + "\n  " + repr(v))
                    c = db.execute(sql, v)
                    reproduced = True

                in_db = list(c.fetchall())
                if in_db:
                    if len(in_db) > 1:
                        m = "WARN: multiple entries: [{}] |{}|\n{}"
                        rep_db = "\n".join([repr(x) for x in in_db])
                        print(m.format(rp, len(in_db), rep_db))
                    continue

                n_added += 1
                wark = hashlib.sha512((rd + "/" + fn).encode("utf-8")).digest()
                wark = base64.urlsafe_b64encode(wark)[:43]

                sql = "insert into up values (?,?,?,?,?)"
                v = (wark, ts, sz, rd, fn)
                try:
                    db.execute(sql, v)
                except Exception as ex:
                    print("insert: " + repr(ex) + "\n  " + repr(v))
                    db.execute(sql, v)
                    reproduced = True
        
        db.commit()
        print("visited {}, added {} entries in total, {}".format(
            n_total, n_added, "REPRODUCED" if reproduced else "passed"))

        if reproduced:
            sys.exit(1)

    def _open_db(self, db_path):
        conn = sqlite3.connect(db_path)
        try:
            nfiles = next(conn.execute("select count(w) from up"))[0]
            print("found DB at {} ({} entries)".format(db_path, nfiles))
            return conn
        except:
            pass

        for cmd in [
            r"create table up (w text, mt int, sz int, rd text, fn text)",
            r"create index up_rd on up(rd)",
            r"create index up_fn on up(fn)",
        ]:
            conn.execute(cmd)

        print("created DB at {}".format(db_path))
        return conn


if __name__ == "__main__":
    if len(sys.argv) > 1:
        make_test_data(int(sys.argv[1]))
    else:
        Up2k().build_indexes()


"""
=======================================================================
sample output:  initial run  (create testdata, create database)
=======================================================================

$ pypy3 ./pypy3-sqlite-bug.py
creating 99890 lines of test data in up2k.txt.gz
0
5000
10000
15000
[...]
85000
90000
95000
created DB at up2k.db
5000
10000
15000
20000
select: InterfaceError('Error binding parameter 0 - probably unsupported type.',)
  ('LYDY46AZJ4BAIQER6UYFUJIU3WQEP5XTLYDY46AZJ4BAIQER6UYFUJIU3WQEP5XT', 'UJIU3WQEP5XT')
25000
30000
35000
[...]
85000
90000
95000
visited 99890, added 99890 new files in total, REPRODUCED
"""


"""
=======================================================================
sample output:  second run  (compare testdata to database values)
=======================================================================

$ pypy3 ./pypy3-sqlite-bug.py
found DB at up2k.db |99890|
5000
10000
15000
20000
25000
30000
WARN: multiple entries: [QUMIVCPVUG5X67EUV735X3IPINRU2Z4TQUMIVCPVUG5X67EUV735X3IPINRU2Z4T/X3IPINRU2Z4T] |2|
(b'VW8xQFnGREKqUuCNcb4_piwIx9gDZm64Gq2x2cshbkC', 1229804881, None, None, None)
(b'VW8xQFnGREKqUuCNcb4_piwIx9gDZm64Gq2x2cshbkC', 1229804881, 1448100694, 'QUMIVCPVUG5X67EUV735X3IPINRU2Z4TQUMIVCPVUG5X67EUV735X3IPINRU2Z4T', 'X3IPINRU2Z4T')
select: InterfaceError('Error binding parameter 0 - probably unsupported type.',)
  ('2PHUVYORDIQ7UUDM33HWJGKWIX2WC7X22PHUVYORDIQ7UUDM33HWJGKWIX2WC7X2', 'JGKWIX2WC7X2')
35000
40000
45000
[...]
85000
90000
95000
visited 99890, added 0 new files in total, REPRODUCED
"""
